
#include "connection.hpp"
#include "event.hpp"
#include "ircmsg.hpp"
#include "linebuffer.hpp"
#include "settings.hpp"
#include "write_irc.hpp"

#include <boost/asio.hpp>

#include <algorithm>
#include <chrono>
#include <fstream>
#include <coroutine>
#include <iostream>
#include <limits>
#include <list>
#include <memory>
#include <string>
#include <string_view>
#include <tuple>
#include <utility>
#include <variant>
#include <vector>
#include <unordered_map>
#include <unordered_set>
#include <coroutine>

#include "ping_thread.hpp"
#include "registration_thread.hpp"
#include "self_thread.hpp"
#include "snote_thread.hpp"

using namespace std::chrono_literals;

struct irc_promise;

struct irc_coroutine : std::coroutine_handle<irc_promise> {
    using promise_type = irc_promise;
};

struct irc_promise {
    std::exception_ptr exception_;

    irc_coroutine get_return_object() { return {irc_coroutine::from_promise(*this)}; }
    std::suspend_never initial_suspend() noexcept { return {}; }
    std::suspend_always final_suspend() noexcept { return {}; }
    void return_void() {}
    void unhandled_exception() {
        exception_ = std::current_exception();
    }
};

struct wait_command {
    Connection& connection_;
    IrcCommand want_cmd_;

    const IrcMsg *result;
    boost::signals2::connection ircmsg_connection_;
    boost::signals2::connection disconnect_connection_;

    wait_command(Connection& connection, IrcCommand want_cmd)
        : connection_{connection}, want_cmd_{want_cmd} {}

    bool await_ready() noexcept { return false; }
    void await_suspend(std::coroutine_handle<irc_promise> handle) {
        ircmsg_connection_ = connection_.sig_ircmsg.connect([this, handle](auto cmd, auto &msg) {
            if (cmd == want_cmd_) {
                ircmsg_connection_.disconnect();
                disconnect_connection_.disconnect();
                result = &msg;
                handle.resume();
            }
        });
        disconnect_connection_ = connection_.sig_disconnect.connect([this, handle]() {
            ircmsg_connection_.disconnect();
            disconnect_connection_.disconnect();
            handle.destroy(); // XXX
        });
    }
    const IrcMsg &await_resume() { return *result; }
};

irc_coroutine example(Connection& connection) {
        auto & msg1 = co_await wait_command {connection, IrcCommand::RPL_WELCOME};
        std::cout << "WELCOME " << msg1.args[0] << "\n";
        auto & msg5 = co_await wait_command {connection, IrcCommand::RPL_ISUPPORT};
        std::cout << "ISUPPORT " << msg5.args[0] << "\n";
}

auto start(boost::asio::io_context & io, Settings const& settings) -> void
{
    auto const connection = std::make_shared<Connection>(io);

    RegistrationThread::start(*connection, settings.password, settings.username, settings.realname, settings.nickname);
    PingThread::start(*connection);
    SelfThread::start(*connection);
    auto const snote_thread = SnoteThread::start(*connection);

    /*
    snote_thread->sig_snote.connect([](auto tag, auto &match) {
        std::cout << "SNOTE " << static_cast<int>(tag) << std::endl;
        for (auto c : match.get_results())
        {
            std::cout << "    " << std::string_view{c.first, c.second} << std::endl;
        }
    });
    */
    auto logic = example(*connection);

    boost::asio::co_spawn(
        io,
        connection->connect(io, settings.host, settings.service),
        [&io, &settings](std::exception_ptr e)
        {
            auto timer = std::make_shared<boost::asio::steady_timer>(io);

            timer->expires_after(5s);
            timer->async_wait([&io, &settings, timer](auto) {
                start(io, settings);
            });
        });
}

auto get_settings() -> Settings
{
    if (auto config_stream = std::ifstream {"config.toml"})
    {
        return Settings::from_stream(config_stream);
    }
    else
    {
        std::cerr << "Unable to open config.toml\n";
        std::exit(1);
    }
}

auto main() -> int
{
    auto const settings = get_settings();
    auto io = boost::asio::io_context{};
    start(io, settings);
    io.run();
}
